//+------------------------------------------------------------------+
//|                                                 OrderManager.mqh |
//|                                      Copyright  2016, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//|   MQL's OOP notes: Singleton, Command queue, and Template method |
//|                                 patterns in simple order manager |
//|                        https://www.mql5.com/en/blogs/post/681732 |
//+------------------------------------------------------------------+

// command queue [https://en.wikipedia.org/wiki/Command_pattern]
// singleton [https://en.wikipedia.org/wiki/Singleton_pattern]
// template method [https://en.wikipedia.org/wiki/Template_method_pattern]

#property strict


#define D2S(A) DoubleToString((A), 6)
#define CURRENT_PRICE 0.0


class OrderCommand
{
  protected:
    static int HandledErrorCodes[];

    static int cmdIncrementedID;
    
    // command general properties
    int cmdID;
    datetime created;
  
    // possible override of manager's settings
    int deviation;
    int magicId;

    // order properties
    string symbol;
    int cmd;
    double volume;
    double price;
    double stoploss;
    double takeprofit;
    string comment;
    datetime expiration;
    color arrow_color;
    
    // auxiliary stuff
    int _LastErrorOverride;
    bool selfDestruction;
    int result;
    
    
    int handledError(string opName, int errCode, bool &priceRequote)
    {
      int handled = -1;
      priceRequote = false;
      for(int i = 0; i < ArraySize(HandledErrorCodes); i++)
      {
        if(HandledErrorCodes[i] == errCode)
        {
          handled = i;
          priceRequote = (i >= 7);
          break;
        }
      }
      Print(opName + " error: ", errCode, " handled:", (handled != -1));
      return handled;
    }
    
  public:
    
    OrderCommand(string symbol = "", int cmd = -1, double volume = 0, double price = 0, int deviation = 0, double stoploss = 0, double takeprofit = 0,
             string comment = "", int magicId = 0, datetime expiration = 0, color arrow_color = CLR_NONE, bool selfDestruction = false)
    {
      this.symbol = symbol == "" || symbol == NULL ? _Symbol : symbol;
      int d = (int)MarketInfo(this.symbol, MODE_DIGITS);
      
      this.cmd = cmd;
      this.volume = volume;
      this.price = NormalizeDouble(price, d);
      this.stoploss = NormalizeDouble(stoploss, d);
      this.takeprofit = NormalizeDouble(takeprofit, d);
      this.expiration = expiration;
      this.comment = comment;
      
      this.deviation = deviation > 0 ? deviation : OrderManager::getInstance().getDeviation();
      this.magicId = magicId > 0 ? magicId : OrderManager::getInstance().getMagic();
      
      this.selfDestruction = selfDestruction;
      
      cmdID = cmdIncrementedID++;
      created = TimeCurrent();
    }

    
    int getCmdID() const
    {
      return cmdID;
    }
    
    static int getNextCmdID()
    {
      return cmdIncrementedID;
    }
    
    datetime getCreated() const
    {
      return created;
    }
    
    string getSymbol() const
    {
      return symbol;
    }

    int getType() const
    {
      return cmd;
    }
    
    double getVolume() const
    {
      return volume;
    }
    
    virtual string toString() const
    {
      return (string)cmdID + " " + classname() + " " + (string)cmd + " " + symbol + " " + (string)volume;
    }

    
    void resetLastError()
    {
      ResetLastError();
      _LastErrorOverride = 0;
    }
    
    int getLastError()
    {
      if(_LastErrorOverride != 0) return _LastErrorOverride;
      int e = GetLastError();
      if(e != 0)
      {
        _LastErrorOverride = e;
      }
      return _LastErrorOverride;
    }
    
    
    // "template method" design pattern
    int process(const int maxRetries, const int timeWaitMsecs)
    {
      int i;
      for(i = 0; i < maxRetries; i++)
      {
        resetLastError();
        
        int code = execute();
        if(code > 0) return code; // success
        
        Print("DATA:", symbol, " ", cmd, " p=", D2S(price), " sl=", D2S(stoploss), " tp=", D2S(takeprofit),
        " / ask=", D2S(MarketInfo(symbol, MODE_ASK)), " bid=", D2S(MarketInfo(symbol, MODE_BID)), " V=", D2S(volume),
        " s=", (int)MarketInfo(symbol, MODE_SPREAD), " ", (int)MarketInfo(symbol, MODE_STOPLEVEL));
        
        int err = getLastError();
        bool priceRequote = false;
        
        if(handledError(classname(), err, priceRequote) == -1) break;

        Sleep(timeWaitMsecs); // NB: doesn't work in the tester
        RefreshRates();
        
        if(cmd == OP_BUY) price = MarketInfo(symbol, MODE_ASK);
        else if(cmd == OP_SELL) price = MarketInfo(symbol, MODE_BID);
        
        // we can't deduce updated price for pending orders and order deletion
        if((cmd > OP_SELL && priceRequote) || cmd < 0) break;
        
        Print("New price=", D2S(price));
      }
      
      if(i == maxRetries)
      {
        Print("Max retries count ", maxRetries, " reached, give up");
        return 0; // neutral, retry later
      }
   
      return -1;
    }
    
    virtual int execute() // = 0;
    {
      _LastErrorOverride = ERR_NO_RESULT;
      return -1;
    }
    
    virtual string classname() const // = 0;
    {
      return typename(this);
    }
    
    int getResult() const
    {
      return result;
    }
    
    bool isSelfDestructing()
    {
      return selfDestruction && (CheckPointer(&this) == POINTER_DYNAMIC);
    }
    
};

class OrderCreateCommand: public OrderCommand
{
  protected:
    // this method provides additional info about possible problems
    bool checkOrder(const int type, const string symbol, double &lot)
    {
      if(IsTesting())
      {
        if(StringCompare(symbol, Symbol()) != 0)
        {
          // TODO: log order to file
          return false;
        }
      }

      double maxLot = MarketInfo(symbol, MODE_MAXLOT);
      if(lot > maxLot) lot = maxLot;
    
      double oneLot = MarketInfo(symbol, MODE_MARGINREQUIRED);
      if(oneLot == 0)
      {
        Print("Required margin = 0 ", symbol);
        return(false);
      }
      double minLot = MarketInfo(symbol, MODE_MINLOT);
      if(lot < minLot)
      {
        Print("Lot ", lot, " is less than allowed ", minLot, ":", symbol);
        return(false);
      }
      
      if(oneLot * lot > AccountFreeMargin())
      {
        Print("Not enough free margin for lot size:", lot);
        return(false);
      }
    
      return(true);
    }
  
  public:
    OrderCreateCommand(int cmd, double volume, string symbol, double price = 0, int deviation = 0, double stoploss = 0, double takeprofit = 0,
      string comment = "", int magicId = 0, datetime expiration = 0, color arrow_color = CLR_NONE, bool selfDestruction = false):
      OrderCommand(symbol, cmd, volume, price, deviation, stoploss, takeprofit, comment, magicId, expiration, arrow_color, selfDestruction)
    {
    }
    
    OrderCreateCommand(int cmd, double volume, string symbol, double price = 0, double stoploss = 0, double takeprofit = 0,
      string comment = "", datetime expiration = 0, bool selfDestruction = false):
      OrderCommand(symbol, cmd, volume, price, 0, stoploss, takeprofit, comment, 0, expiration, CLR_NONE)
    {
    }
    
    OrderCreateCommand(int cmd, double volume):
      OrderCommand(_Symbol, cmd, volume)
    {
      selfDestruction = CheckPointer(&this) == POINTER_DYNAMIC; // convenient but tricky implicit initialization
    }
    
    // builder pattern begins
    OrderCreateCommand *price(const double price)
    {
      this.price = price;
      return &this;
    }

    OrderCreateCommand *stoploss(const double stoploss)
    {
      this.stoploss = stoploss;
      return &this;
    }

    OrderCreateCommand *takeprofit(const double takeprofit)
    {
      this.takeprofit = takeprofit;
      return &this;
    }

    OrderCreateCommand *comment(const string comment)
    {
      this.comment = comment;
      return &this;
    }
    
    OrderCreateCommand *symbol(const string symbol)
    {
      this.symbol = symbol;
      return &this;
    }
    
    OrderCreateCommand *expiration(const datetime expiration)
    {
      this.expiration = expiration;
      return &this;
    }
    
    // concrete implementation of the command's main task
    virtual int execute()
    {
      if(!checkOrder(cmd, symbol, volume))
      {
        _LastErrorOverride = ERR_TRADE_ERROR;
        return -1;
      }
      
      if(price == 0)
      {
        if(cmd == OP_BUY) price = MarketInfo(symbol, MODE_ASK);
        else if(cmd == OP_SELL) price = MarketInfo(symbol, MODE_BID);
      }
      
      result = OrderSend(symbol, cmd, volume, price, deviation, stoploss, takeprofit, comment, magicId, expiration, arrow_color);
      return result;
    }
    
    virtual string classname() const
    {
      return typename(this);
    }
};

class OrderCreateCommandWithDelay: public OrderCreateCommand
{
  private:
    int delay;
    
  public:
    OrderCreateCommandWithDelay(int delay, int cmd, double volume, string symbol, double price = 0, int deviation = 0, double stoploss = 0, double takeprofit = 0,
      string comment = "", int magicId = 0, datetime expiration = 0, color arrow_color = CLR_NONE, bool selfDestruction = false):
      OrderCreateCommand(cmd, volume, symbol, price, deviation, stoploss, takeprofit, comment, magicId, expiration, arrow_color, selfDestruction)
    {
      this.delay = delay;
    }

    OrderCreateCommandWithDelay(int delay, int cmd, double volume):
      OrderCreateCommand(cmd, volume)
    {
      this.delay = delay;
      if(delay > 0) Print("delay till ", (string)(created + delay)); // debug output
    }
    
    virtual int execute()
    {
      if(created + delay > TimeCurrent())
      {
        Print("delay"); // debug output
        // randomly switch between handled (ERR_BROKER_BUSY) and unhandled (ERR_MARKET_CLOSED) simulated error
        this._LastErrorOverride = MathRand() > 32767 / 2 ? ERR_MARKET_CLOSED : ERR_BROKER_BUSY;
        return -1;
      }

      if(price == 0)
      {
        if(cmd == OP_BUY) price = MarketInfo(symbol, MODE_ASK);
        else if(cmd == OP_SELL) price = MarketInfo(symbol, MODE_BID);
      }
      
      result = OrderSend(symbol, cmd, volume, price, deviation, stoploss, takeprofit, comment, magicId, expiration, arrow_color);
      return result;
    }

    virtual string classname() const
    {
      return typename(this);
    }
};

class OrderModifyCommand: public OrderCommand
{
  private:
    int ticket;
    
  public:
    OrderModifyCommand(int ticket, double price = 0, double stoploss = 0, double takeprofit = 0,
      datetime expiration = 0, color arrow_color = CLR_NONE, bool selfDestruction = false):
      OrderCommand("", -1, 0, price, 0, stoploss, takeprofit, NULL, 0, expiration, arrow_color, selfDestruction)
    {
      this.ticket = ticket;
      if(OrderSelect(ticket, SELECT_BY_TICKET))
      {
        if(this.price == 0) this.price = OrderOpenPrice();
        this.symbol = OrderSymbol();
        this.cmd = OrderType();
        this.volume = OrderLots();
        this.comment = OrderComment();
        this.magicId = OrderMagicNumber();
      }
    }
    
    virtual int execute()
    {
      result = OrderModify(ticket, price, stoploss, takeprofit, expiration, arrow_color);
      return result;
    }
    
    virtual string classname() const
    {
      return typename(this);
    }
};

class OrderCloseByCommand: public OrderCommand
{
  private:
    int ticket;
    int other;
    
  public:
    OrderCloseByCommand(int ticket, int opposite, color arrow = CLR_NONE, bool selfDestruction = false):
      OrderCommand("", -1, 0, 0, 0, 0, 0, NULL, 0, 0, arrow, selfDestruction)
    {
      this.ticket = ticket;
      this.other = opposite;
    }
    
    virtual int execute()
    {
      result = OrderCloseBy(ticket, other, arrow_color);
      return result;
    }
    
    virtual string classname() const
    {
      return typename(this);
    }
};

class OrderCloseCommand: public OrderCommand
{
  private:
    int ticket;
    
  public:
    OrderCloseCommand(int ticket, double lots, double price, int deviation = 0, color arrow = CLR_NONE, bool selfDestruction = false):
      OrderCommand("", -1, lots, price, deviation, 0, 0, NULL, 0, 0, arrow, selfDestruction)
    {
      this.ticket = ticket;
      if(OrderSelect(ticket, SELECT_BY_TICKET))
      {
        // exchange Ask and Bid for correct price for closing
        int type = OrderType();
        this.cmd = type == OP_BUY ? OP_SELL : (type == OP_SELL ? OP_BUY : -1); 
        this.symbol = OrderSymbol();
        this.comment = OrderComment();
        this.magicId = OrderMagicNumber();
        
        if(lots == 0)
        {
          this.volume = OrderLots();
        }
      }
    }
    
    virtual int execute()
    {
      if(price == 0) // use current price
      {              
        if(cmd == OP_BUY) price = MarketInfo(symbol, MODE_ASK);
        else if(cmd == OP_SELL) price = MarketInfo(symbol, MODE_BID);
      }
      
      result = OrderClose(ticket, volume, price, deviation, arrow_color);
      return result;
    }

    virtual string classname() const
    {
      return typename(this);
    }
};

class OrderDeleteCommand: public OrderCommand
{
  private:
    int ticket;
    
  public:
    OrderDeleteCommand(int ticket, color arrow = CLR_NONE, bool selfDestruction = false):
      OrderCommand("", -1, 0, 0, 0, 0, 0, NULL, 0, 0, arrow, selfDestruction)
    {
      this.ticket = ticket;
    }
    
    virtual int execute()
    {
      result = OrderDelete(ticket, arrow_color);
      return result;
    }

    virtual string classname() const
    {
      return typename(this);
    }
};

class OrderCallback
{
  public:
    virtual void onSuccess(const int result, const OrderCommand *cmd)    // = 0;
    {
      Print("Completed ", cmd.toString(), ", result: ", result);
    }
    virtual void onFailure(const int error, const OrderCommand *cmd)     // = 0;
    {
      Print("Failed ", cmd.toString(), ", error: ", error);
    }
    virtual void onTimeout(const int timeout, const OrderCommand *cmd)   // = 0;
    {
      Print("Timeout ", cmd.toString(), ", timeout: ", timeout);
    }
};


class OrderManager
{
  private:

    int slippage;
    int magic;
  
    int maxRetries;
    int timeWaitMsecs;
    
    OrderCommand *commands[]; // "command" queue design pattern
    
    static OrderManager *manager; // singleton object instance
    
    OrderCallback *pCallback;
    

    void revoke(const int i)
    {
      if(CheckPointer(commands[i]) == POINTER_DYNAMIC && commands[i].isSelfDestructing())
      {
        delete commands[i];
      }
      commands[i] = NULL;
    }

    void revokeAll()
    {
      const int n = ArraySize(commands);
      for(int i = 0; i < n; i++)
      {
        revoke(i);
      }
      ArrayResize(commands, 0);
    }
    
    // try to process the command right away
    // on success or failure return result
    // on temporary error place command to the queue
    int async(OrderCommand *cmd)
    {
      int result = cmd.process(maxRetries, timeWaitMsecs);
      if(result != 0)
      {
        if(cmd.isSelfDestructing())
        {
          delete cmd;
        }
        return result;
      }
      place(cmd);
      return 0; // neutral, retry later
    }
    
    OrderManager() // constructor is hidden from outer space
    {
      maxRetries = 10;
      timeWaitMsecs = 5000;
      slippage = 10;
      
      manager = GetPointer(this);
      
      pCallback = NULL;
    }

  public:

    // "singleton" design pattern
    static OrderManager *getInstance()
    {
      if(manager == NULL) manager = new OrderManager();
      return manager;
    }
    
    // add a command to the queue
    void place(OrderCommand *cmd)
    {
      int i, n = ArraySize(commands);
      for(i = 0; i < n; i++)
      {
        if(commands[i] == NULL) break;
      }
      
      if(i == n)
      {
        ArrayResize(commands, n + 1);
      }
      else
      {
        n = i;
      }
      commands[n] = cmd;
    }
    
    // process all pending commands in order of appearance
    void processAll()
    {
      int n = ArraySize(commands);
      for(int i = 0; i < n; i++)
      {
        if(CheckPointer(commands[i]) != POINTER_INVALID)
        {
          int r = commands[i].process(maxRetries, timeWaitMsecs);
          if(r != 0) // succeeded or failed, otherwise 0 - still pending
          {
            if(CheckPointer(pCallback) != POINTER_INVALID)
            {
              if(r > 0)
              {
                pCallback.onSuccess(r, commands[i]);
              }
              else // r < 0
              {
                pCallback.onFailure(commands[i].getLastError(), commands[i]);
              }
            }

            revoke(i);
          }
        }
      }
    }
    
    // get allocated size of the command array
    int getSize() const
    {
      return ArraySize(commands);
    }

    // get number of pending commands
    int getActiveSize() const
    {
      int size = 0;
      int n = ArraySize(commands);
      for(int i = 0; i < n; i++)
      {
        if(CheckPointer(commands[i]) != POINTER_INVALID)
        {
          size++;
        }
      }
      return size;
    }
    
    // get command by index in the array
    OrderCommand *getCommand(const int i)
    {
      if(i >= 0 && i < ArraySize(commands))
      {
        return commands[i];
      }
      return NULL;
    }

    // return last added command
    OrderCommand *getLastCommand()
    {
      int n = ArraySize(commands);
      for(int i = 0; i < n; i++)
      {
        if(CheckPointer(commands[i]) != POINTER_INVALID && commands[i].getCmdID() == (OrderCommand::getNextCmdID() - 1))
        {
          return commands[i];
        }
      }
      
      return NULL;
    }
    
    // remove a queued command by ID
    void remove(const int cmdId)
    {
      int n = ArraySize(commands);
      for(int i = 0; i < n; i++)
      {
        if(CheckPointer(commands[i]) != POINTER_INVALID)
        {
          if(commands[i].getCmdID() == cmdId)
          {
            revoke(i);
          }
        }
      }
    }
    
    // remove too old pending commands
    // which are still in queue after specified number of seconds after creation
    void deleteByTimeout(const int seconds)
    {
      int n = ArraySize(commands);
      for(int i = 0; i < n; i++)
      {
        if(CheckPointer(commands[i]) != POINTER_INVALID)
        {
          if(commands[i].getCreated() + seconds < TimeCurrent())
          {
            if(CheckPointer(pCallback) != POINTER_INVALID)
            {
              pCallback.onTimeout(seconds, commands[i]);
            }
            revoke(i);
          }
        }
      }
    }
    
    // set or unset (if NULL passed) an object with notification handlers
    void bindCallback(OrderCallback *c)
    {
      pCallback = c;
    }
    

    //   E V E N T S
    
    void onTimer()
    {
      processAll();
    }
    
    void onInit()
    {
      // TODO: restore command queue from Global Variables
    }
    
    void onDeinit()
    {
      // TODO: serialize pending commands to Global Variables
      revokeAll();
      if(CheckPointer(&this) == POINTER_DYNAMIC) delete &this;
    }
    

    //   S E T T E R S
    
    OrderManager *setDeviation(int d)
    {
      slippage = d;
      return GetPointer(this);
    }

    OrderManager *setMagic(int m)
    {
      magic = m;
      return GetPointer(this);
    }
    
    OrderManager *setMaxRetries(int number)
    {
      maxRetries = number;
      return GetPointer(this);
    }
    
    OrderManager *setTimeWait(int milliseconds)
    {
      timeWaitMsecs = milliseconds;
      return GetPointer(this);
    }
    

    //   G E T T E R S
    
    int getDeviation() const
    {
      return slippage;
    }
    
    int getMaxRetries() const
    {
      return maxRetries;
    }
    
    int getTimeWait() const
    {
      return timeWaitMsecs;
    }
    
    int getMagic() const
    {
      return magic;
    }
    
    
    
    //   S H O R T H A N D S
    
    int Send(string symbol, int cmd, double volume, double price, double stoploss, double takeprofit,
             string comment = "", datetime expiration = 0, color arrow_color = CLR_NONE)
    {
      return Send(symbol, cmd, volume, price, slippage, stoploss, takeprofit, comment, magic, expiration, arrow_color);
    }
    
    int Send(string symbol, int cmd, double volume, double price, int deviation, double stoploss, double takeprofit,
             string comment = "", int magicId = 0, datetime expiration = 0, color arrow_color = CLR_NONE)
    {
      OrderCreateCommand create(cmd, volume, symbol, price, deviation, stoploss, takeprofit, comment, magicId, expiration, arrow_color);
      return create.process(maxRetries, timeWaitMsecs);
    }
    
    int SendAsync(string symbol, int cmd, double volume, double price, int deviation, double stoploss, double takeprofit,
             string comment = "", int magicId = 0, datetime expiration = 0, color arrow_color = CLR_NONE)
    {
      return async(new OrderCreateCommand(cmd, volume, symbol, price, deviation, stoploss, takeprofit, comment, magicId, expiration, arrow_color, true));
    }
    
    bool Modify(int ticket, double price, double stoploss, double takeprofit,
                datetime expiration, color arrow_color = CLR_NONE)
    {
      OrderModifyCommand modify(ticket, price, stoploss, takeprofit, expiration, arrow_color);
      return modify.process(maxRetries, timeWaitMsecs);
    }

    bool ModifyAsync(int ticket, double price, double stoploss, double takeprofit,
                datetime expiration, color arrow_color = CLR_NONE)
    {
      return async(new OrderModifyCommand(ticket, price, stoploss, takeprofit, expiration, arrow_color, true));
    }
    
    bool CloseBy(int ticket, int opposite, color arrow_color = CLR_NONE)
    {
      OrderCloseByCommand closeby(ticket, opposite, arrow_color);
      return closeby.process(maxRetries, timeWaitMsecs);
    }

    bool CloseByAsync(int ticket, int opposite, color arrow_color = CLR_NONE)
    {
      return async(new OrderCloseByCommand(ticket, opposite, arrow_color, true));
    }

    bool Close(int ticket, double lots, double price, color Color = CLR_NONE)
    {
      return Close(ticket, lots, price, slippage, Color);
    }
    
    bool Close(int ticket, double lots, double price, int deviation, color arrow_color = CLR_NONE)
    {
      OrderCloseCommand close(ticket, lots, price, deviation, arrow_color);
      return close.process(maxRetries, timeWaitMsecs);
    }

    bool CloseAsync(int ticket, double lots, double price, int deviation, color arrow_color = CLR_NONE)
    {
      return async(new OrderCloseCommand(ticket, lots, price, deviation, arrow_color, true));
    }
    
    bool Delete(int ticket, color arrow_color = CLR_NONE)
    {
      OrderDeleteCommand d(ticket, arrow_color);
      return d.process(maxRetries, timeWaitMsecs);
    }
    
    bool DeleteAsync(int ticket, color arrow_color = CLR_NONE)
    {
      return async(new OrderDeleteCommand(ticket, arrow_color, true));
    }
};

OrderManager *OrderManager::manager = NULL; // first call to OrderManager::getInstance() create this (if not yet exist)

int OrderCommand::HandledErrorCodes[] = {0, ERR_SERVER_BUSY, ERR_NO_CONNECTION, ERR_TRADE_TIMEOUT, ERR_OFF_QUOTES, ERR_BROKER_BUSY, ERR_TRADE_CONTEXT_BUSY,
                           ERR_INVALID_PRICE, ERR_PRICE_CHANGED, ERR_REQUOTE};
int OrderCommand::cmdIncrementedID = 1;

#ifndef ASYNC_ORDERS

#define OrderSend OrderManager::getInstance().Send
#define OrderModify OrderManager::getInstance().Modify
#define OrderClose OrderManager::getInstance().Close
#define OrderCloseBy OrderManager::getInstance().CloseBy
#define OrderDelete OrderManager::getInstance().Delete

#else

#define OrderSend OrderManager::getInstance().SendAsync
#define OrderModify OrderManager::getInstance().ModifyAsync
#define OrderClose OrderManager::getInstance().CloseAsync
#define OrderCloseBy OrderManager::getInstance().CloseByAsync
#define OrderDelete OrderManager::getInstance().DeleteAsync

#endif


